# Copyright 2017 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = '1.2.0'
__author__ = 'Francesco Ceccon'
__author_email__ = 'francesco@ceccon.me'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright 2018 Francesco Ceccon'
__url__ = 'https://github.com/cog-imperial/suspect'
__description__ = 'Special Structure Detection for Pyomo'
